import os
from base64 import b64encode
from io import BytesIO
from urllib.parse import quote as urlquote
import sys
from kaleido.scopes.plotly import PlotlyScope
import logging


class Plot(object):
    """A Plot object is defined by a path to the output file and the title of the plot."""

    only_report = False
    
    def __init__(self, path, title):
        self.path = path
        self.title = title
        self.fig = None
        self.html = None

    def encode(self):
        if self.html:
            return self.html
        elif self.fig:
            return self.encode2()
        else:
            return self.encode1()

    def encode1(self):
        """Return the base64 encoding of the figure file and insert in html image tag."""
        data_uri = b64encode(open(self.path, "rb").read()).decode("utf-8").replace("\n", "")
        return '<img src="data:image/png;base64,{0}">'.format(data_uri)

    def encode2(self):
        """Return the base64 encoding of the fig attribute and insert in html image tag."""
        buf = BytesIO()
        self.fig.savefig(buf, format="png", bbox_inches="tight", dpi=100)
        buf.seek(0)
        string = b64encode(buf.read())
        return '<img src="data:image/png;base64,{0}">'.format(urlquote(string))

    def save(self, settings):
        if not(self.only_report):
            if self.html:
                with open(self.path, "w") as html_out:
                    html_out.write(self.html)
                if not settings["no_static"]:
                    try:
                        for fmt in settings["format"]:
                            self.save_static(fmt)
                    except (AttributeError, ValueError) as e:
                        p = os.path.splitext(self.path)[0] + ".png"
                        if os.path.exists(p):
                            os.remove(p)

                        logging.warning("No static plots are saved due to some kaleido problem:")
                        logging.warning(e)

            elif self.fig:
                # if settings["format"] is a list, save the figure in all formats
                if isinstance(settings["format"], list):
                    for fmt in settings["format"]:
                        self.fig.savefig(
                            fname=self.path + "." + fmt,
                            format=fmt,
                            bbox_inches="tight",
                        )
                else:
                    self.fig.savefig(fname=self.path, format=settings["format"], bbox_inches="tight")
            else:
                sys.exit("No method to save plot object: no html or fig defined.")

    def show(self):
        if self.fig:
            return self.fig.fig
        else:
            sys.stderr.write(".show not implemented for Plot instance without fig attribute!")

    def save_static(self, figformat):
        scope = PlotlyScope()
        with open(self.path.replace("html", figformat), "wb") as f:
            f.write(scope.transform(self.fig, format=figformat))
            logging.info(
                f"Saved {self.path.replace('.html', '')}  as {figformat} (or png for --legacy)"
            )



import sys
import os
from datetime import datetime as dt
from time import time
import logging
from nanoplot.version import __version__
from argparse import HelpFormatter, Action, ArgumentParser
import textwrap as _textwrap
import pandas as pd


class CustomHelpFormatter(HelpFormatter):
    def _format_action_invocation(self, action):
        if not action.option_strings or action.nargs == 0:
            return super()._format_action_invocation(action)
        default = self._get_default_metavar_for_optional(action)
        args_string = self._format_args(action, default)
        return ", ".join(action.option_strings) + " " + args_string

    def _fill_text(self, text, width, indent):
        return "".join(indent + line for line in text.splitlines(keepends=True))

    def _split_lines(self, text, width):
        text = self._whitespace_matcher.sub(" ", text).strip()
        return _textwrap.wrap(text, 80)


class Action_Print_Colors(Action):
    def __init__(self, option_strings, dest="==SUPPRESS==", default="==SUPPRESS==", help=None):
        super(Action_Print_Colors, self).__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        list_colors()


class Action_Print_Colormaps(Action):
    def __init__(self, option_strings, dest="==SUPPRESS==", default="==SUPPRESS==", help=None):
        super(Action_Print_Colormaps, self).__init__(
            option_strings=option_strings, dest=dest, default=default, nargs=0, help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        list_colormaps()


def get_args():
    epilog = """EXAMPLES:
    NanoPlot --summary sequencing_summary.txt --loglength -o summary-plots-log-transformed
    NanoPlot -t 2 --fastq reads1.fastq.gz reads2.fastq.gz --maxlength 40000 --plots hex dot
    NanoPlot --color yellow --bam alignment1.bam alignment2.bam alignment3.bam --downsample 10000
    """
    parser = ArgumentParser(
        description="Creates various plots for long read sequencing data.".upper(),
        epilog=epilog,
        formatter_class=custom_formatter,
        add_help=False,
    )
    general = parser.add_argument_group(title="General options")
    general.add_argument("-h", "--help", action="help", help="show the help and exit")
    general.add_argument(
        "-v",
        "--version",
        help="Print version and exit.",
        action="version",
        version="NanoPlot {}".format(__version__),
    )
    general.add_argument(
        "-t",
        "--threads",
        help="Set the allowed number of threads to be used by the script",
        default=4,
        type=int,
    )
    general.add_argument(
        "--verbose", help="Write log messages also to terminal.", action="store_true"
    )
    general.add_argument(
        "--store",
        help="Store the extracted data in a pickle file for future plotting.",
        action="store_true",
    )
    general.add_argument(
        "--raw", help="Store the extracted data in tab separated file.", action="store_true"
    )
    general.add_argument("--huge", help="Input data is one very large file.", action="store_true")
    general.add_argument(
        "-o", "--outdir", help="Specify directory in which output has to be created.", default="."
    )
    general.add_argument("--no_static", help="Do not make static (png) plots.", action="store_true")
    general.add_argument(
        "-p",
        "--prefix",
        help="Specify an optional prefix to be used for the output files.",
        default="",
        type=str,
    )
    general.add_argument(
        "--tsv_stats",
        help="Output the stats file as a properly formatted TSV.",
        action="store_true",
    )
    general.add_argument(
        "--only-report",
        help="Output only the report",
        action="store_true",
        default=False,
    )
    general.add_argument(
        "--info_in_report", help="Add NanoPlot run info in the report.", action="store_true"
    )
    filtering = parser.add_argument_group(
        title="Options for filtering or transforming input prior to plotting"
    )
    filtering.add_argument(
        "--maxlength", help="Hide reads longer than length specified.", type=int, metavar="N"
    )
    filtering.add_argument(
        "--minlength", help="Hide reads shorter than length specified.", type=int, metavar="N"
    )
    filtering.add_argument(
        "--drop_outliers", help="Drop outlier reads with extreme long length.", action="store_true"
    )
    filtering.add_argument(
        "--downsample", help="Reduce dataset to N reads by random sampling.", type=int, metavar="N"
    )
    filtering.add_argument(
        "--loglength",
        help="Additionally show logarithmic scaling of lengths in plots.",
        action="store_true",
    )
    filtering.add_argument(
        "--percentqual",
        help="Use qualities as theoretical percent identities.",
        action="store_true",
    )
    filtering.add_argument(
        "--alength",
        help="Use aligned read lengths rather than sequenced length (bam mode)",
        action="store_true",
    )
    filtering.add_argument(
        "--minqual",
        help="Drop reads with an average quality lower than specified.",
        type=int,
        metavar="N",
    )
    filtering.add_argument(
        "--runtime_until", help="Only take the N first hours of a run", type=int, metavar="N"
    )
    filtering.add_argument(
        "--readtype",
        help="Which read type to extract information about from summary. \
                                 Options are 1D, 2D, 1D2",
        default="1D",
        choices=["1D", "2D", "1D2"],
    )
    filtering.add_argument(
        "--barcoded",
        help="Use if you want to split the summary file by barcode",
        action="store_true",
    )
    filtering.add_argument(
        "--no_supplementary",
        help="Use if you want to remove supplementary alignments",
        action="store_true",
        default=False,
    )
    visual = parser.add_argument_group(title="Options for customizing the plots created")
    visual.add_argument(
        "-c", "--color", help="Specify a valid matplotlib color for the plots", default="#4CB391"
    )
    visual.add_argument(
        "-cm",
        "--colormap",
        help="Specify a valid matplotlib colormap for the heatmap",
        default="Greens",
    )
    visual.add_argument(
        "-f",
        "--format",
        help="Specify the output format of the plots, which are in addition to the html files",
        default=["png"],
        nargs="*",
        choices=["png", "jpg", "jpeg", "webp", "svg", "pdf", "eps", "json"],
    )
    visual.add_argument(
        "--plots",
        help="Specify which bivariate plots have to be made.",
        default=["kde", "dot"],
        type=str,
        nargs="*",
        choices=["kde", "hex", "dot"],
    )
    visual.add_argument(
        "--legacy",
        help="Specify which bivariate plots have to be made (legacy mode).",
        type=str,
        nargs="*",
        choices=["kde", "dot", "hex"],
    )
    visual.add_argument(
        "--listcolors",
        help="List the colors which are available for plotting and exit.",
        action=Action_Print_Colors,
        default=False,
    )
    visual.add_argument(
        "--listcolormaps",
        help="List the colors which are available for plotting and exit.",
        action=Action_Print_Colormaps,
        default=False,
    )
    visual.add_argument(
        "--no-N50", help="Hide the N50 mark in the read length histogram", action="store_true"
    )
    visual.add_argument(
        "--N50",
        help="Show the N50 mark in the read length histogram",
        action="store_true",
        default=False,
    )
    visual.add_argument(
        "--title",
        help="Add a title to all plots, requires quoting if using spaces",
        type=str,
        default=None,
    )
    visual.add_argument(
        "--font_scale", help="Scale the font of the plots by a factor", type=float, default=1
    )
    visual.add_argument("--dpi", help="Set the dpi for saving images", type=int, default=100)
    visual.add_argument(
        "--hide_stats",
        help="Not adding Pearson R stats in some bivariate plots",
        action="store_true",
        default=False,
    )
    target = parser.add_argument_group(title="Input data sources, one of these is required.")
    mtarget = target.add_mutually_exclusive_group(required=True)
    mtarget.add_argument(
        "--fastq", help="Data is in one or more default fastq file(s).", nargs="+", metavar="file"
    )
    mtarget.add_argument(
        "--fasta", help="Data is in one or more fasta file(s).", nargs="+", metavar="file"
    )
    mtarget.add_argument(
        "--fastq_rich",
        help="Data is in one or more fastq file(s) generated by albacore, \
                               MinKNOW or guppy with additional information \
                               concerning channel and time.",
        nargs="+",
        metavar="file",
    )
    mtarget.add_argument(
        "--fastq_minimal",
        help="Data is in one or more fastq file(s) generated by albacore, \
                               MinKNOW or guppy with additional information concerning channel \
                               and time. Is extracted swiftly without elaborate checks.",
        nargs="+",
        metavar="file",
    )
    mtarget.add_argument(
        "--summary",
        help="Data is in one or more summary file(s) generated by albacore \
                               or guppy.",
        nargs="+",
        metavar="file",
    )
    mtarget.add_argument(
        "--bam", help="Data is in one or more sorted bam file(s).", nargs="+", metavar="file"
    )
    mtarget.add_argument(
        "--ubam", help="Data is in one or more unmapped bam file(s).", nargs="+", metavar="file"
    )
    mtarget.add_argument(
        "--cram", help="Data is in one or more sorted cram file(s).", nargs="+", metavar="file"
    )
    mtarget.add_argument("--pickle", help="Data is a pickle file stored earlier.", metavar="pickle")
    mtarget.add_argument(
        "--feather",
        "--arrow",
        help="Data is in one or more feather file(s).",
        nargs="+",
        metavar="file",
    )
    args = parser.parse_args()
    if args.listcolors:
        list_colors()
    if args.listcolormaps:
        list_colormaps()
    if args.no_N50:
        sys.stderr.write("DeprecationWarning: --no-N50 is currently the default setting.\n")
        sys.stderr.write("The argument is thus unnecessary but kept for backwards compatibility.")
    if args.barcoded and not args.summary:
        sys.exit("ARGUMENT ERROR: --barcoded only works with data provided as --summary!")
    settings = vars(args)
    settings["path"] = os.path.join(args.outdir, args.prefix)
    return settings, args


def custom_formatter(prog):
    return CustomHelpFormatter(prog)


def list_colors():
    parent_directory = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
    colours = open(os.path.join(parent_directory, "extra/color_options_hex.txt"))
    col_hex = {}

    for line in colours:
        key, value = line.split(",")
        col_hex[key] = value.strip()
    print("Valid colors: {}".format("\n".join([c.strip() for c in list(col_hex.keys())])))
    sys.exit(0)


def list_colormaps():
    print(
        "Valid colormaps:\nGreys\nYlGnBu\nGreens\nYlOrRd\nBluered\nRdBu\nReds\nBlues\nPicnic\n"
        "Rainbow\nPortland\nJet\nHot\nBlackbody\nEarth\nElectric\nViridis\nCividis"
    )
    sys.exit(0)


def make_output_dir(path):
    try:
        if not os.path.exists(path):
            os.makedirs(path)
    except IOError:
        sys.exit("ERROR: No writing permission to the output directory.")


def init_logs(args, tool="NanoPlot"):
    """Initiate log file and log arguments."""
    start_time = dt.fromtimestamp(time()).strftime("%Y%m%d_%H%M")
    logname = os.path.join(args.outdir, args.prefix + tool + "_" + start_time + ".log")
    handlers = [logging.FileHandler(logname)]
    if args.verbose:
        handlers.append(logging.StreamHandler())
    logging.basicConfig(format="%(asctime)s %(message)s", handlers=handlers, level=logging.INFO)
    logging.info("{} {} started with arguments {}".format(tool, __version__, args))
    logging.info("Python version is: {}".format(sys.version.replace("\n", " ")))
    return logname


def subsample_datasets(df, minimal=10000):
    if "dataset" in df:
        list_df = []

        for d in df["dataset"].unique():
            dataset = df.loc[df["dataset"] == d]

            if len(dataset.index) < minimal:
                list_df.append(dataset)

            else:
                list_df.append(dataset.sample(minimal))

        subsampled_df = pd.concat(list_df, ignore_index=True)

    else:
        if len(df.index) < minimal:
            subsampled_df = df

        else:
            subsampled_df = df.sample(minimal)

    return subsampled_df



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
exec(open("nanoplot/version.py").read())

setup(
    name="NanoPlot",
    version=__version__,  # noqa: F821
    description="Plotting suite for Oxford Nanopore sequencing data and alignments",
    long_description=open(path.join(here, "README.md"), encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/wdecoster/NanoPlot",
    author="Wouter De Coster",
    author_email="decosterwouter@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    keywords="nanopore sequencing plotting quality control",
    packages=find_packages() + ["extra", "scripts"],
    python_requires=">=3",
    install_requires=[
        "biopython",
        "pysam>0.10.0.0",
        "pandas>=1.1.0",
        "numpy>=1.16.5",
        "scipy",
        "python-dateutil",
        "nanoget>=1.19.1",
        "nanomath>=1.0.0",
        "plotly>=5.4.0",
        "pyarrow",
        "kaleido",
    ],
    package_data={"NanoPlot": []},
    package_dir={"nanoplot": "nanoplot"},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "NanoPlot=nanoplot.NanoPlot:main",
        ],
    },
    data_files=[("", ["LICENSE"])],
)



# wdecoster
"""
This module provides functions for plotting data extracted from Oxford Nanopore sequencing
reads and alignments, but some of it's functions can also be used for other applications.


FUNCTIONS
* Check if a specified color is a valid matplotlib color
check_valid_color(color)
* Check if a specified output format is valid
checkvalidFormat(format)
* Create a bivariate plot with dots, hexbins and/or kernel density estimates.
Also arguments for specifying axis names, color and xlim/ylim
scatter(x, y, names, path, color, format, plots, stat=None, log=False, minvalx=0, minvaly=0)
* Create cumulative yield plot and evaluate read length and quality over time
timePlots(df, path, color, format)
* Create length distribution histogram and density curve
lengthPlots(array, name, path, n50, color, format, log=False)
* Create flowcell physical layout in numpy array
makeLayout()
* Present the activity (number of reads) per channel on the flowcell as a heatmap
spatialHeatmap(array, title, path, color, format)

"""

import plotly.graph_objs as go
import plotly
import logging
import sys
import os
import pandas as pd
import numpy as np
from nanoplotter.plot import Plot
import plotly.express as px
import plotly.figure_factory as ff
from nanoplotter.spatial_heatmap import spatial_heatmap
from nanoplotter.timeplots import time_plots
import re
import nanoplot.utils as utils


def check_valid_color(color):
    """Check if the color provided by the user is valid.

    If color is invalid the default is returned.
    """
    colors, _ = colors_and_colormaps()
    if color in colors:
        logging.info(f"NanoPlot:  Valid color {color}.")
        return colors.get(color)

    elif re.search(r"^#(?:[0-9a-fA-F]{3}){1,2}$", color):
        logging.info(f"NanoPlot:  Valid color {color}.")
        return color

    else:
        logging.info(f"NanoPlot:  Invalid color {color}, using default.")
        sys.stderr.write("Invalid color {color}, using default.\n")
        return "#4CB391"


def check_valid_colormap(colormap):
    """Check if the colormap provided by the user is valid.

    If colormap is invalid the default is returned.
    """
    _, colormaps = colors_and_colormaps()
    if colormap in colormaps:
        logging.info(f"NanoPlot:  Valid colormap {colormap}.")
        return colormap
    else:
        logging.info(f"NanoPlot:  Invalid colormap {colormap}, using default.")
        sys.stderr.write(f"Invalid colormap {colormap}, using default.\n")
        return "Greens"


def scatter(
    x,
    y,
    legacy,
    names,
    path,
    plots,
    color,
    colormap,
    settings,
    stat=None,
    log=False,
    minvalx=0,
    minvaly=0,
    title=None,
    xmax=None,
    ymax=None,
):
    """->
    create marginalised scatterplots and KDE plot with marginalized histograms
    -> update from scatter_legacy function to utilise plotly package
    - scatterplot with histogram on both axes
    - kernel density plot with histograms on both axes
    - hexbin not implemented yet
    - pauvre plot temporarily not available
    """
    logging.info(f"NanoPlot: Creating {names[0]} vs {names[1]} plots using {x.size} reads.")
    if not contains_variance([x, y], names):
        return []
    plots_made = []
    idx = np.random.choice(x.index, min(10000, len(x)), replace=False)
    maxvalx = xmax or np.amax(x[idx])
    maxvaly = ymax or np.amax(y[idx])

    if plots["dot"]:
        if log:
            dot_plot = Plot(
                path=path + "_loglength_dot.html",
                title=f"{names[0]} vs {names[1]} plot using dots "
                "after log transformation of read lengths",
            )
        else:
            dot_plot = Plot(
                path=path + "_dot.html", title=f"{names[0]} vs {names[1]} plot using dots"
            )

        fig = px.scatter(
            x=x[idx],
            y=y[idx],
            marginal_x="histogram",
            marginal_y="histogram",
            range_x=[minvalx, maxvalx],
            range_y=[minvaly, maxvaly],
        )
        fig.update_traces(marker=dict(color=color))
        fig.update_yaxes(rangemode="tozero")
        fig.update_xaxes(rangemode="tozero")

        fig.update_layout(
            xaxis_title=names[0], yaxis_title=names[1], title=title or dot_plot.title, title_x=0.5
        )

        if log:
            ticks = [10**i for i in range(10) if not 10**i > 10 * (10**maxvalx)]
            fig.update_layout(
                xaxis=dict(tickmode="array", tickvals=np.log10(ticks), ticktext=ticks, tickangle=45)
            )

        dot_plot.fig = fig
        dot_plot.html = dot_plot.fig.to_html(full_html=False, include_plotlyjs="cdn")
        dot_plot.save(settings)
        plots_made.append(dot_plot)

    if plots["kde"]:
        kde_plot = Plot(
            path=path + "_loglength_kde.html" if log else path + "_kde.html",
            title=f"{names[0]} vs {names[1]} kde plot",
        )

        col = hex_to_rgb_scale_0_1(color)
        fig = ff.create_2d_density(
            x[idx], y[idx], point_size=3, hist_color=col, point_color=col, colorscale=colormap
        )

        fig.update_layout(
            xaxis_title=names[0],
            yaxis_title=names[1],
            title=title or kde_plot.title,
            title_x=0.5,
            xaxis=dict(tickangle=45),
        )

        if log:
            ticks = [10**i for i in range(10) if not 10**i > 10 * (10**maxvalx)]
            fig.update_layout(
                xaxis=dict(tickmode="array", tickvals=np.log10(ticks), ticktext=ticks, tickangle=45)
            )

        kde_plot.fig = fig
        kde_plot.html = kde_plot.fig.to_html(full_html=False, include_plotlyjs="cdn")
        kde_plot.save(settings)
        plots_made.append(kde_plot)

    if 1 in legacy.values():
        settings, args = utils.get_args()
        plots_made += scatter_legacy(
            x=x[idx],
            y=y[idx],
            names=names,
            path=path,
            plots=legacy,
            color=color,
            settings=settings,
            stat=stat,
            log=log,
            minvalx=minvalx,
            minvaly=minvaly,
            title=title,
        )
    return plots_made


def scatter_legacy(
    x,
    y,
    names,
    path,
    plots,
    color,
    settings,
    stat=None,
    log=False,
    minvalx=0,
    minvaly=0,
    title=None,
    xmax=None,
    ymax=None,
):
    """Create bivariate plots.

    Create four types of bivariate plots of x vs y, containing marginal summaries
    -A scatter plot with histograms on axes
    -A hexagonal binned plot with histograms on axes
    -A kernel density plot with density curves on axes
    -A pauvre-style plot using code from https://github.com/conchoecia/pauvre
    """
    try:
        import matplotlib as mpl

        mpl.use("Agg")
        import seaborn as sns
        import matplotlib.pyplot as plt
    except ImportError:
        sys.stderr.write("NanoPlot needs seaborn and matplotlib with --legacy")
        return []
    figformat = settings["format"]
    if figformat in ["webp", "json"]:
        figformat = "png"

    logging.info(f"NanoPlot: Creating {names[0]} vs {names[1]} legacy plots using {x.size} reads.")
    if not contains_variance([x, y], names):
        return []
    sns.set(style="ticks")
    maxvalx = xmax or np.amax(x)
    maxvaly = ymax or np.amax(y)

    plots_made = []
    path = path + "_legacy"

    if plots["hex"]:
        if log:
            hex_plot = Plot(
                path=path + "_loglength_hex." + figformat[0],
                title="{} vs {} plot using hexagonal bins "
                "after log transformation of read lengths".format(names[0], names[1]),
            )
        else:
            hex_plot = Plot(
                path=path + "_hex." + figformat[0],
                title="{} vs {} plot using hexagonal bins".format(names[0], names[1]),
            )
        plot = sns.jointplot(
            x=x,
            y=y,
            kind="hex",
            color=color,
            stat_func=stat,
            space=0,
            xlim=(minvalx, maxvalx),
            ylim=(minvaly, maxvaly),
            height=10,
        )
        plot.set_axis_labels(names[0], names[1])
        if log:
            ticks = [10**i for i in range(10) if not 10**i > 10 * (10**maxvalx)]
            plot.ax_joint.set_xticks(np.log10(ticks))
            plot.ax_marg_x.set_xticks(np.log10(ticks))
            plot.ax_joint.set_xticklabels(ticks)
        plt.subplots_adjust(top=0.90)
        plot.fig.suptitle(title or f"{names[0]} vs {names[1]} plot", fontsize=25)
        hex_plot.fig = plot
        hex_plot.save(settings)
        plots_made.append(hex_plot)

    sns.set(style="darkgrid")
    if plots["dot"]:
        print("we here")
        if log:
            dot_plot = Plot(
                path=path + "_loglength_dot." + figformat[0],
                title="{} vs {} plot using dots "
                "after log transformation of read lengths".format(names[0], names[1]),
            )
        else:
            dot_plot = Plot(
                path=path + "_dot." + figformat[0],
                title="{} vs {} plot using dots".format(names[0], names[1]),
            )
        plot = sns.jointplot(
            x=x,
            y=y,
            kind="scatter",
            color=color,
            stat_func=stat,
            xlim=(minvalx, maxvalx),
            ylim=(minvaly, maxvaly),
            space=0,
            height=10,
            joint_kws={"s": 1},
        )
        plot.set_axis_labels(names[0], names[1])
        if log:
            ticks = [10**i for i in range(10) if not 10**i > 10 * (10**maxvalx)]
            plot.ax_joint.set_xticks(np.log10(ticks))
            plot.ax_marg_x.set_xticks(np.log10(ticks))
            plot.ax_joint.set_xticklabels(ticks)
        plt.subplots_adjust(top=0.90)
        plot.fig.suptitle(title or "{} vs {} plot".format(names[0], names[1]), fontsize=25)
        dot_plot.fig = plot
        dot_plot.save(settings)
        plots_made.append(dot_plot)

    if plots["kde"]:
        if len(x) > 2:
            idx = np.random.choice(x.index, min(2000, len(x)), replace=False)
            if log:
                kde_plot = Plot(
                    path=path + "_loglength_kde." + figformat[0],
                    title="{} vs {} plot using a kernel density estimation "
                    "after log transformation of read lengths".format(names[0], names[1]),
                )
            else:
                kde_plot = Plot(
                    path=path + "_kde." + figformat[0],
                    title=f"{names[0]} vs {names[1]} plot using a kernel density estimation",
                )
            plot = sns.jointplot(
                x=x[idx],
                y=y[idx],
                kind="kde",
                clip=((0, np.Inf), (0, np.Inf)),
                xlim=(minvalx, maxvalx),
                ylim=(minvaly, maxvaly),
                space=0,
                color=color,
                stat_func=stat,
                shade_lowest=False,
                height=10,
            )
            plot.set_axis_labels(names[0], names[1])
            if log:
                ticks = [10**i for i in range(10) if not 10**i > 10 * (10**maxvalx)]
                plot.ax_joint.set_xticks(np.log10(ticks))
                plot.ax_marg_x.set_xticks(np.log10(ticks))
                plot.ax_joint.set_xticklabels(ticks)
            plt.subplots_adjust(top=0.90)
            plot.fig.suptitle(title or "{} vs {} plot".format(names[0], names[1]), fontsize=25)
            kde_plot.fig = plot
            kde_plot.save(settings)
            plots_made.append(kde_plot)
        else:
            sys.stderr.write("Not enough observations (reads) to create a kde plot.\n")
            logging.info("NanoPlot: Not enough observations (reads) to create a kde plot")
    plt.close("all")
    return plots_made


# def pauvre_plot():
#     from pauvre.marginplot import margin_plot
#     if plots["pauvre"] and names == ['Read lengths', 'Average read quality'] and log is False:
#         pauvre_plot = Plot(
#             path=path + "_pauvre." + figformat,
#             title="{} vs {} plot using pauvre-style @conchoecia".format(names[0], names[1]))
#         sns.set(style="white")
#         margin_plot(df=pd.DataFrame({"length": x, "meanQual": y}),
#                     Y_AXES=False,
#                     title=title or "Length vs Quality in Pauvre-style",
#                     plot_maxlen=None,
#                     plot_minlen=0,
#                     plot_maxqual=None,
#                     plot_minqual=0,
#                     lengthbin=None,
#                     qualbin=None,
#                     BASENAME="whatever",
#                     path=pauvre_plot.path,
#                     fileform=[figformat],
#                     dpi=600,
#                     TRANSPARENT=True,
#                     QUIET=True)
#         plots_made.append(pauvre_plot)


def contains_variance(arrays, names):
    """
    Make sure both arrays for bivariate ("scatter") plot have a stddev > 0
    """
    for ar, name in zip(arrays, names):
        if np.std(ar) == 0:
            sys.stderr.write(f"No variation in '{name.lower()}', skipping bivariate plots.\n")
            logging.info(f"NanoPlot: No variation in {name}, skipping bivariate plot")
            return False
    else:
        return True


def length_plots(array, name, path, settings, title=None, n50=None, color="#4CB391"):
    """Create histogram of normal and log transformed read lengths."""
    logging.info("NanoPlot:  Creating length plots for {}.".format(name))
    maxvalx = np.amax(array)
    if n50:
        logging.info(
            "NanoPlot: Using {} reads with read length N50 of {}bp and maximum of {}bp.".format(
                array.size, n50, maxvalx
            )
        )
    else:
        logging.info(f"NanoPlot:  Using {array.size} reads maximum of {maxvalx}bp.")

    plots = []

    HistType = [
        {"weight": array, "name": "Weighted", "ylabel": "Number of bases"},
        {"weight": None, "name": "Non weighted", "ylabel": "Number of reads"},
    ]

    for h_type in HistType:
        histogram = Plot(
            path=path
            + h_type["name"].replace(" ", "_")
            + "Histogram"
            + name.replace(" ", "")
            + ".html",
            title=f"{h_type['name']} histogram of read lengths",
        )

        hist, bin_edges = np.histogram(
            array, bins=max(round(int(maxvalx) / 500), 10), weights=h_type["weight"]
        )

        fig = go.Figure()

        fig.add_trace(go.Bar(x=bin_edges[1:], y=hist, marker_color=color))

        if n50:
            fig.add_vline(n50)
            fig.add_annotation(text="N50", x=n50, y=0.95)
            fig.update_annotations(font_size=8)

        fig.update_layout(
            xaxis_title="Read length",
            yaxis_title=h_type["ylabel"],
            title=title or histogram.title,
            title_x=0.5,
        )

        histogram.fig = fig
        histogram.html = histogram.fig.to_html(full_html=False, include_plotlyjs="cdn")
        histogram.save(settings)

        log_histogram = Plot(
            path=path
            + h_type["name"].replace(" ", "_")
            + "LogTransformed_Histogram"
            + name.replace(" ", "")
            + ".html",
            title=h_type["name"] + " histogram of read lengths after log transformation",
        )

        if h_type["weight"] is None:
            hist_log, bin_edges_log = np.histogram(
                np.log10(array), bins=max(round(int(maxvalx) / 500), 10), weights=h_type["weight"]
            )

        else:
            hist_log, bin_edges_log = np.histogram(
                np.log10(array),
                bins=max(round(int(maxvalx) / 500), 10),
                weights=np.log10(h_type["weight"]),
            )

        fig = go.Figure()
        fig.add_trace(
            go.Bar(
                x=bin_edges_log[1:],
                y=hist_log,
                marker_color=color,
                hovertext=["{:.0f}".format(10**i) for i in bin_edges_log[1:]],
                hoverinfo="y+text",
            )
        )

        ticks = [10**i for i in range(10) if not 10**i > 10 * maxvalx]

        fig.update_layout(
            xaxis=dict(tickmode="array", tickvals=np.log10(ticks), ticktext=ticks),
            xaxis_title="Read length",
            yaxis_title=h_type["ylabel"],
            title=title or log_histogram.title,
            title_x=0.5,
        )

        if n50:
            fig.add_vline(np.log10(n50))
            fig.add_annotation(text="N50", x=np.log10(n50), y=0.95)
            fig.update_annotations(font_size=8)

        log_histogram.fig = fig
        log_histogram.html = log_histogram.fig.to_html(full_html=False, include_plotlyjs="cdn")
        log_histogram.save(settings)

        plots.extend([histogram, log_histogram])

    plots.append(
        yield_by_minimal_length_plot(
            array=array, name=name, path=path, title=title, color=color, settings=settings
        )
    )

    return plots


def dynamic_histogram(array, name, path, settings, title=None, color="#4CB391"):
    """
    Use plotly to a histogram
    Return html code, but also save as png
    """
    dynhist = Plot(
        path=path + f"Dynamic_Histogram_{name[0].lower() + name[1:].replace(' ', '_')}.html",
        title="Dynamic histogram of {}".format(name[0].lower() + name[1:]),
    )
    ylabel = "Number of reads" if len(array) <= 10000 else "Downsampled number of reads"
    dynhist.html, dynhist.fig = plotly_histogram(
        array=array.sample(min(len(array), 10000)),
        color=color,
        title=title or dynhist.title,
        xlabel=name,
        ylabel=ylabel,
    )
    dynhist.save(settings)
    return dynhist


def plotly_histogram(array, color="#4CB391", title=None, xlabel=None, ylabel=None):
    data = [go.Histogram(x=array, opacity=0.4, marker=dict(color=color))]
    html = plotly.offline.plot(
        {
            "data": data,
            "layout": go.Layout(
                barmode="overlay", title=title, yaxis_title=ylabel, xaxis_title=xlabel
            ),
        },
        output_type="div",
        show_link=False,
    )
    fig = go.Figure(
        {"data": data, "layout": go.Layout(barmode="overlay", title=title, title_x=0.5)}
    )
    return html, fig


def yield_by_minimal_length_plot(array, name, path, settings, title=None, color="#4CB391"):
    df = pd.DataFrame(data={"lengths": np.sort(array)[::-1]})
    df["cumyield_gb"] = df["lengths"].cumsum() / 10**9
    idx = np.random.choice(array.index, min(10000, len(array)), replace=False)

    yield_by_length = Plot(path=path + "Yield_By_Length.html", title="Yield by length")

    fig = px.scatter(df, x=df.reindex(idx)["lengths"], y=df.reindex(idx)["cumyield_gb"])
    fig.update_traces(marker=dict(color=color))
    fig.update_layout(
        xaxis_title="Read length",
        yaxis_title="Cumulative yield for minimal length [Gb]",
        title=title or yield_by_length.title,
        title_x=0.5,
    )

    yield_by_length.fig = fig
    yield_by_length.html = yield_by_length.fig.to_html(full_html=False, include_plotlyjs="cdn")
    yield_by_length.save(settings)

    return yield_by_length


def colors_and_colormaps():
    colormaps = (
        "Greys,YlGnBu,Greens,YlOrRd,Bluered,RdBu,Reds,Blues,Picnic,Rainbow,Portland,Jet,"
        "Hot,Blackbody,Earth,Electric,Viridis,Cividis"
    ).split(",")
    parent_directory = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
    colours = open(os.path.join(parent_directory, "extra/color_options_hex.txt"))
    col_hex = {}

    for line in colours:
        key, value = line.split(",")
        col_hex[key] = value.strip()

    return col_hex, colormaps


def hex_to_rgb_scale_0_1(hexcolor):
    color = hexcolor.lstrip("#")
    RGB_color = tuple(int(color[x : x + 2], 16) for x in (0, 2, 4))

    RGB_color = [x / 255 for x in RGB_color]

    return tuple(RGB_color)


def run_tests():
    import pickle

    df = pickle.load(open("nanotest/sequencing_summary.pickle", "rb"))
    scatter(
        x=df["lengths"],
        y=df["quals"],
        names=["Read lengths", "Average read quality"],
        path="LengthvsQualityScatterPlot",
        plots={"dot": 1, "kde": 1},
    )
    time_plots(df=df, path="./", color="#4CB391")
    length_plots(array=df["lengths"], name="lengths", path="./")
    spatial_heatmap(
        array=df["channelIDs"],
        title="Number of reads generated per channel",
        path="ActivityMap_ReadsPerChannel",
    )


if __name__ == "__main__":
    run_tests()
